from typing import List

from bpy.types import (
    Material,
    ShaderNodeGroup,
)

import animation as anim
from pyedm_platform_selector import pyedm
from objects_custom_props import get_edm_props
from serializer import SLink
from materials.material_interface import IMaterial
from materials.material_edm_node_groups import NODE_GROUP_GLASS_NAME
from materials.material_wrap import (
    MaterialWrap, 
    TextureDesk,
    ValueDesk,
)
from enums import (
    BpyNodeSocketType,
    NodeSocketCommonEnum,
    NodeSocketInGlassEnum,
)
from materials.materials_common import (
    make_def_base_block,
    make_def_rmo_block,
    make_def_normal_block,
    make_def_flir_block,
    make_def_damage_block,
    make_def_bone_block,
    create_blocks,
    get_shadow_caster_value,
    get_transparency_value,
    get_glass_value
)


GLASS_MATERIAL_NAME = 'EDM_Glass_Material'

class GlassMaterialWrap(MaterialWrap):
    class Textures:
        def __init__(self, node_group: ShaderNodeGroup) -> None:
            self.albedo         = TextureDesk(node_group, NodeSocketInGlassEnum.BASE_COLOR, BpyNodeSocketType.COLOR)
            self.decal          = TextureDesk(node_group, NodeSocketInGlassEnum.DECAL_COLOR, BpyNodeSocketType.COLOR)
            self.rmo            = TextureDesk(node_group, NodeSocketInGlassEnum.ROUGH_METAL, BpyNodeSocketType.COLOR)
            self.light_map      = TextureDesk(node_group, NodeSocketInGlassEnum.LIGHTMAP, BpyNodeSocketType.COLOR)
            self.normal         = TextureDesk(node_group, NodeSocketInGlassEnum.NORMAL, BpyNodeSocketType.COLOR)
            self.flir           = TextureDesk(node_group, NodeSocketInGlassEnum.FLIR, BpyNodeSocketType.COLOR)
            self.damage_color   = TextureDesk(node_group, NodeSocketInGlassEnum.DAMAGE_COLOR, BpyNodeSocketType.COLOR)
            self.damage_normal  = TextureDesk(node_group, NodeSocketInGlassEnum.DAMAGE_NORMAL, BpyNodeSocketType.COLOR)
            self.damage_mask    = TextureDesk(node_group, NodeSocketInGlassEnum.DAMAGE_MASK, BpyNodeSocketType.COLOR)
            self.damage_alpha   = TextureDesk(node_group, NodeSocketInGlassEnum.DAMAGE_ALPHA, BpyNodeSocketType.COLOR)

    class Values:
        def __init__(self, node_group: ShaderNodeGroup) -> None:
            self.alpha              = ValueDesk(node_group, NodeSocketInGlassEnum.BASE_ALPHA, BpyNodeSocketType.FLOAT, 1.0)
            self.decal_alpha        = ValueDesk(node_group, NodeSocketInGlassEnum.DECAL_ALPHA, BpyNodeSocketType.FLOAT, 1.0)
            self.light_map_value    = ValueDesk(node_group, NodeSocketInGlassEnum.LIGHTMAP_VALUE, BpyNodeSocketType.FLOAT, 1.0)
            self.blend_mode         = ValueDesk(node_group, NodeSocketInGlassEnum.TRANSPARENCY, BpyNodeSocketType.TRANSPARENCY, 'OPAQUE')
            self.shadow_caster      = ValueDesk(node_group, NodeSocketInGlassEnum.SHADOW_CASTER, BpyNodeSocketType.SHADOWCASTER, 'SHADOW_CASTER_YES')
            self.glass_type         = ValueDesk(node_group, NodeSocketInGlassEnum.GLASS_TYPE, BpyNodeSocketType.GLASSTYPE, 'GLASS_INSTRUMENTAL')
            self.version            = ValueDesk(node_group, NodeSocketCommonEnum.VERSION, BpyNodeSocketType.INTEGER, -1)
            self.opacity_value      = ValueDesk(node_group, NodeSocketInGlassEnum.OPACITY_VALUE, BpyNodeSocketType.FLOAT, 1.0)

    def __init__(self, material: Material) -> None:
        super().__init__(material, GLASS_MATERIAL_NAME, 'BSDF')
        self.textures: GlassMaterialWrap.Textures = None
        self.values: GlassMaterialWrap.Values = None
        self.apply_group(self.node_group, GlassMaterialWrap.Textures, GlassMaterialWrap.Values)        
        if not self.valid:
            return
        self.update_linked_resources()


## Default material
class GlassMaterial(IMaterial):
    name = GLASS_MATERIAL_NAME
    description_file_name = f'data/{GLASS_MATERIAL_NAME}.pickle'
    factory = GlassMaterialWrap
    node_group_name = NODE_GROUP_GLASS_NAME
    
    # TODO-270 - refactor
    # class Textures:
    #     def __init__(self, node_group: ShaderNodeGroup) -> None:
    #         self.albedo         = TextureDesk(node_group, NodeSocketInGlassEnum.BASE_COLOR, BpyNodeSocketType.COLOR)
    #         self.decal          = TextureDesk(node_group, NodeSocketInGlassEnum.DECAL_COLOR, BpyNodeSocketType.COLOR)
    #         self.rmo            = TextureDesk(node_group, NodeSocketInGlassEnum.ROUGH_METAL, BpyNodeSocketType.COLOR)
    #         self.light_map      = TextureDesk(node_group, NodeSocketInGlassEnum.LIGHTMAP, BpyNodeSocketType.COLOR)
    #         self.normal         = TextureDesk(node_group, NodeSocketInGlassEnum.NORMAL, BpyNodeSocketType.COLOR)
    #         self.flir           = TextureDesk(node_group, NodeSocketInGlassEnum.FLIR, BpyNodeSocketType.COLOR)
    #         self.damage_color   = TextureDesk(node_group, NodeSocketInGlassEnum.DAMAGE_COLOR, BpyNodeSocketType.COLOR)
    #         self.damage_normal  = TextureDesk(node_group, NodeSocketInGlassEnum.DAMAGE_NORMAL, BpyNodeSocketType.COLOR)
    #         self.damage_mask    = TextureDesk(node_group, NodeSocketInGlassEnum.DAMAGE_MASK, BpyNodeSocketType.COLOR)
    #         self.damage_alpha   = TextureDesk(node_group, NodeSocketInGlassEnum.DAMAGE_ALPHA, BpyNodeSocketType.COLOR)

    # class Values:
    #     def __init__(self, node_group: ShaderNodeGroup) -> None:
    #         self.alpha              = ValueDesk(node_group, NodeSocketInGlassEnum.BASE_ALPHA, BpyNodeSocketType.FLOAT, 1.0)
    #         self.decal_alpha        = ValueDesk(node_group, NodeSocketInGlassEnum.DECAL_ALPHA, BpyNodeSocketType.FLOAT, 1.0)
    #         self.light_map_value    = ValueDesk(node_group, NodeSocketInGlassEnum.LIGHTMAP_VALUE, BpyNodeSocketType.FLOAT, 1.0)
    #         self.blend_mode         = ValueDesk(node_group, NodeSocketInGlassEnum.TRANSPARENCY, BpyNodeSocketType.TRANSPARENCY, 'OPAQUE')
    #         self.shadow_caster      = ValueDesk(node_group, NodeSocketInGlassEnum.SHADOW_CASTER, BpyNodeSocketType.SHADOWCASTER, 'SHADOW_CASTER_YES')
    #         self.glass_type         = ValueDesk(node_group, NodeSocketInGlassEnum.GLASS_TYPE, BpyNodeSocketType.GLASS_TYPE, 'GLASS_INSTRUMENTAL')
    #         self.version            = ValueDesk(node_group, NodeSocketCommonEnum.VERSION, BpyNodeSocketType.INTEGER, -1)
    #         self.opacity_value      = ValueDesk(node_group, NodeSocketInGlassEnum.OPACITY_VALUE, BpyNodeSocketType.FLOAT, 1.0)


    @classmethod    
    def build_blocks(cls, object, material_wrap: MaterialWrap, mesh_storage):
        edm_props = get_edm_props(object)
        edm_render_node = pyedm.GlassNode(object.name, material_wrap.material.name)        
        edm_render_node.setIndices(mesh_storage.indices)

        block_func = [
            make_def_base_block,
            make_def_rmo_block,
            make_def_normal_block,
            make_def_flir_block,
            make_def_damage_block,
            make_def_bone_block,
        ]
        blocks = create_blocks(mesh_storage, material_wrap, edm_props, block_func)
        for block in blocks:
            edm_render_node.addBlock(block)
        
        transparency_mode: int = get_transparency_value(material_wrap.values.blend_mode.value)
        edm_render_node.setTransparentMode(transparency_mode)

        shadow_caster_type: int = get_shadow_caster_value(material_wrap.values.shadow_caster.value)
        edm_render_node.setShadowCaster(shadow_caster_type)

        glass_type: int = get_glass_value(material_wrap.values.glass_type.value)
        edm_render_node.setGlassType(glass_type)

        opacity_value_path: str = material_wrap.values.opacity_value.anim_path
        is_color_value_animated: bool = opacity_value_path and anim.has_path_anim(material_wrap.material.node_tree.animation_data, opacity_value_path)
        if is_color_value_animated and edm_props.OPACITY_VALUE_ARG != -1:
            arg_opacity_value_n: int = edm_props.OPACITY_VALUE_ARG
            key_opacity_value_list: anim.KeyFramePoints = anim.extract_anim_float(material_wrap.material.node_tree.animation_data.action, opacity_value_path)
            opacity_value_prop = pyedm.PropertyFloat(arg_opacity_value_n, key_opacity_value_list)
            edm_render_node.setOpacityValue(opacity_value_prop)
        else:
            opacity_value_prop = pyedm.PropertyFloat(material_wrap.values.opacity_value.value)
            edm_render_node.setOpacityValue(opacity_value_prop)

        two_sided: bool = edm_props.TWO_SIDED
        edm_render_node.setTwoSided(two_sided)

        return edm_render_node
    
    @classmethod
    def process_links(cls, old_links, old_version, group_node_type_name):
        return old_links
    
    @classmethod 
    def restore_defaults(cls, old_sockest, new_node_group, old_version, material_name):
        pass

